--[[

Changelog
20160908 - 1.5
	Public
		Fixed bug that could cause addon options to reset (for real this time!)
		
	Detailed
		Added check to Initialize to ensure this addon was the source of the ADDON_LOADED event

20160908 - 1.4
	Public
		Added support for Artifact Relics
		Fixed bug whereby sometimes multiple notifications would be given for the same item

	Known Limitations
		Personal Loot Helper cannot determine whether an Artifact Relic is an upgrade for a character's offspec since it has no way of determining what relics are slotted into the offspec Artifact
	
	Detailed
		Added ValidRelics, PLH_GetRelicType, GetEquippedRelic
		Added relic evaluation to IsEquippableItemForCharacter and IsAnUpgradeForCharacter
		Added relics to cache and expected item count
		Checked if frames already exist in Initialize() to prevent creating duplicate listeners

20160903 - 1.3
	Public
		Fixed bug that could result in incorrect recommendations for rings, trinkets, and weapons
		Fixed bug that could cause addon options to reset
		Increased number of group member names that PLH will show a recommendation for from 3 to 4; beyond that, it will state "and others"
		Modified notification text to say "...[item] is an ilvl upgrade..." vs. "...[item] is an upgrade..." to avoid the impression that the addon is doing any type of dps simultation to determine whether an item is truly an upgrade

	Detailed
		Fixed slot IDs to use numbers instead of strings in IsAnUpgradeForCharacter
		Changed Initialize() to be called by an event listener on ADDON_LOADED instead of called directly
		Increased MAX_NAMES_TO_SHOW from 3 to 4
		Commented out some debug statements and calls to PLH_PrintCache to reduce debug spam
		Modified text in PerformNotify
	
20160901 - 1.2
	Public
		Added ability for users to limit evaluations to characters' current spec only (in interface options); by default, PLH will consider an item as a possible upgrade as long as it is usable by ANY of the character's possible specilizations
		Added cache refreshes when players change specs or change equipped gear to ensure evaluations are done with up-to-date spec/gear
		Added support for evaluating whether Legion trinkets are upgrades based on class/spec.  Note that pre-Legion trinkets will not be evaluated.
	
	Detailed
		Added PLH_CURRENT_SPEC_ONLY to IsAnUpgradeForCharacter and interface options screen
		Added PLAYER_SPECIALIZATION_CHANGED listener to update cache when a group member's spec changes
		Added UNIT_INVENTORY_CHANGED listener to update cache when a player's gear changes
		Cleaned up comments to reflect current logic
		Added GetExpectedItemCount to determine how many items should be cached by class/spec instead of assuming 15 for everyone
		Added TRINKET_arrays, IsTrinketUsable, and logic in IsEquippableItemForCharacter to evaluate trinkets

20160830 - 1.1
	Public
		Added logic to limit the upgrade evaluation to gear that is appropriate for the character's possible specs based on the item's primary attribute (str/int/agi)
			Note: will add future enhancement to allow users to select an "evaluate current spec only" option; for now I
			default it to consider items if they're equippable by ANY of the character's specs since players may queue for a spec other than their main spec
		Fixed evaluation of cloaks; previously only considered cloth-wearers as eligible to equip cloaks
		Minor performance improvements, particularly when in a large raid group
		Fixed /plh to open directly to Personal Loot Helper options on first try
		
	Detailed
		In PopulateGroupCache(), only refresh cache if it has been more than 3 seconds since the last refresh
		In IsCharacterInGroup, eliminate redundant call to GetRaidRosterInfo()
		Copy code from BlizzBugsSuck.lua to fix InterfaceOptionsFrame_OpenToCategory not opening to addon
		In UpdateGroupInfoCache(unit), cache the character's Spec, and in PLH_GetItemCountFromCache subtract 2 since we added Spec
		In IsEquippableItemForCharacter, consider the item equippable for every class if it's a cloak
		In IsEquippableItemForCharacter, evaluate the character's spec vs. the primary attribute of the item
		Added PLH_TestItems()
		
The general flow is as follows:
LootReceivedEvent()				triggered when loot is received by anyone in the group
  --> PerformNotify()			if loot is a tradeable and is an upgrade for someone in group, notifies based on users' preferred Notify Mode
								if Coordinate Rolls mode then flow continues as follows:
    --> WhisperReceivedEvent()  triggered when we receive a 'TRADE' whisper from a person who looted a tradeable item
      --> AskForRolls()			prompts the group to roll for the item
        --> PLH_EndRolls()			notifies the group who won the roll

Known Limitations:
	If the user reloads their UI or logs out, all state is lost - the cache will be lost and any pending rolls will remain uncompleted
	If a player receives more than 1 item of personal loot, only their last item will be eligible for rolls when they
	   reply to the addon's request to trade the item(s)
	The addon assumes everyone in the group is eligible to receive tradeable loot; it doesn't check whether everyone
		was within range of the kill and not already loot-locked
	The addon doesn't properly check if tier tokens can be used or are upgrades, but that should be a moot point for
	   Legion since tier will drop as actual gear vs. tier tokens when loot method is personal
	If multiple people are running the addon in Notify Group mode, then each of them will send a message notifying the
	   group of the loot.  There's no straightforward foolproof way around this, so users will have to coordinate
	   among themselves to switch to Notify Self mode if it becomes too spammy.
	Players cannot inspect other characters while the player is dead.  If you die & reload, the item cache will be
	   empty until the next time an event that triggers a cache refresh is received.
	The addon assumes that any loot received is part of the Personal Loot system, if Personal Loot is enabled.  However,
	   items can be obtained other ways - for example, someone using a baleful token to create a baleful item.  Those
	   items are not part of the Personal Loot system, but PLHelper cannot distinguish them from regular loot.
	Trinket evaluation is hardcoded for Legion items only; trinkets prior to Legion will not be evaluated

TODO
	Bug? - it broadcast twice when I received Braided Silver Ring.  Nothing particularly notable about the occassion,
		other than maybe the fact that I also received Jewel of Victory (artifact points).  Could LootReceivedEvent
		have triggered twice for the two loots, yet somehow evaluated Braided Silver Ring on both events??
		
TODO Refactoring
	refactor - get constants for slotid, spec, etc. from _G instead of hard-coding
	refactor - make sure we properly use name/character/unit/guid everywhere to avoid confusion...
			don't even use 'character'; it's confusing; just use name, unit, or guid
	refactor - rename all uses of PLHelper to PLH
	refactor - can probably change queuedRollOwner and queuedRollItem now that I understand lua arrays better; search 'key', for example
	refactor - clean up unit test lua
	refactor - clean up comments; reflect current logic
	refactor - since PLH_GetUnitNameWithRealm() operates on units, can we replace all of its calls with UnitFullName(unit)?
			easy way to test this:  change implementation of PLH_GetUnitNameWithRealm(unit) to call UnitFullName(unit) and see
			what happens!
	comments - update to reflect the approach to looping logic; just using outer loop now; no inner retry logic
  
TOTEST Tests that still need to be run:
	TOTEST:  Legion relics...need to add them to things that reference inventory slots?

Future enhancement ideas:
	Add pop up window giving player option to whisper people who can use gear to offer the loot to them
	add plh trade [item] so user can force rolls (or re-rolls)
	add ability for people to opt out of receiving whispers when they receive tradeable items
	if you have addon installed, get a popup to decide what to do vs. a whisper?
	option for roll timers - auto-finish rolls after [X] seconds or never, timers for roll warnings
	add ability to cancel rolls?
	add localization; see http://lua-users.org/wiki/StringLibraryTutorial for defining string constants
	add a 'Who Else is Running PLHelper?' button to the config screen, which will use SendAddonMessage to query
	   for other instances of PLHelper, and report back the selected Notify Mode of each (to help coordinate with
	   making sure only 1 person is running in Notify Group mode)
	also check for gem slots & tertiary stats when determining if an item is an upgrade
	if Blizz fixes the UIDROPDOWNMENU_MENU_LEVEL taint issue, re-enable the Quality and ILVL dropdowns on the config screen
		and change Notify Mode back into a dropdown.  Taint issue is as follows:
			--8/26 22:33:52.225  Global variable UIDROPDOWNMENU_MENU_LEVEL tainted by PLHelper - Interface\FrameXML\UIDropDownMenu.lua:38
			--8/26 22:33:52.225      securecall()
			--8/26 22:33:52.225      Interface\FrameXML\UIDropDownMenu.lua:64 UIDropDownMenu_Initialize()
			--8/26 22:33:52.225      Interface\AddOns\PLHelper\PLHelperConfig.lua:63
			--"There are a few solutions. The simplest should be to treat UIDROPDOWNMENU_MENU_LEVEL in the same manner as UIDROPDOWNMENU_INIT_MENU/UIDROPDOWNMENU_OPEN_MENU and set it through a secure frame and SetAttribute rather than directly. This should be a quick, guaranteed solution to the issue, ensuring that the variable will always remain secure (unless the user mucks with it directly, which should not ever be necessary)."

]]--

-- slash command
SLASH_PLHelperCommand1 = '/plh'

local NOTIFY_MODE_SELF = 1
local NOTIFY_MODE_GROUP = 2
local NOTIFY_MODE_COORDINATE_ROLLS = 3

local DEFAULT_NOTIFY_MODE = NOTIFY_MODE_GROUP
local DEFAULT_INCLUDE_BOE = false
local DEFAULT_MIN_ILVL = 528  -- personal loot was introduced with Siege of Orgrimmar, which started at ilvl 528
local DEFAULT_MIN_QUALITY = 3  -- Rare
local DEFAULT_DEBUG = false
local DEFAULT_CURRENT_SPEC_ONLY = false

local TRADE_MESSAGE = 'TRADE'
local DELAY_BETWEEN_ROLLS = 4 -- in seconds
local DELAY_BETWEEN_INSPECTIONS = 3  -- in seconds
local MAX_INSPECT_LOOPS = 3    -- maximum # of times to retry calling NotifyInspect on all members in the roster for whom we've cached fewer than the expected number of items
local NUM_EXPECTED_ITEMS = 15 -- number of items we expect each person to have equipped (based on having something in every gear) plus 3 relics
	-- slot; if we've cached fewer than that amount of items for a character, we'll include that character in additional
	-- inspect loops.
--local MAX_INSPECT_RETRIES = 2  -- maximum # of times to retry calling NotifyInspect for a specific character if we don't get an INSPECT_READY
local NUM_EXPECTED_RELICS = 2
local MAX_NAMES_TO_SHOW = 4
local PLH_RELICSLOT = 1000  -- for indexing relics in groupInfoCache

local ValidGear = {
--	{ 'Death Knight', 'Armor', 'Cloth' },
--	{ 'Death Knight', 'Armor', 'Leather' },
--	{ 'Death Knight', 'Armor', 'Mail' },
	{ 'Death Knight', 'Armor', 'Plate' },
	{ 'Death Knight', 'Armor', 'Miscellaneous' },
	{ 'Death Knight', 'Weapon', 'One-Handed Axes' },
	{ 'Death Knight', 'Weapon', 'One-Handed Maces' },
	{ 'Death Knight', 'Weapon', 'Polearms' },
	{ 'Death Knight', 'Weapon', 'One-Handed Swords' },
	{ 'Death Knight', 'Weapon', 'Two-Handed Axes' },
	{ 'Death Knight', 'Weapon', 'Two-Handed Maces' },
	{ 'Death Knight', 'Weapon', 'Two-Handed Swords' },

--	{ 'Demon Hunter', 'Armor', 'Cloth' },
	{ 'Demon Hunter', 'Armor', 'Leather' },
	{ 'Demon Hunter', 'Armor', 'Miscellaneous' },
	{ 'Demon Hunter', 'Weapon', 'One-Handed Axes' },
	{ 'Demon Hunter', 'Weapon', 'Daggers' },
	{ 'Demon Hunter', 'Weapon', 'Fist Weapons' },
	{ 'Demon Hunter', 'Weapon', 'One-Handed Maces' },
	{ 'Demon Hunter', 'Weapon', 'One-Handed Swords' },
	{ 'Demon Hunter', 'Weapon', 'Warglaives' },

--	{ 'Druid', 'Armor', 'Cloth' },
	{ 'Druid', 'Armor', 'Leather' },
	{ 'Druid', 'Armor', 'Miscellaneous' },
	{ 'Druid', 'Weapon', 'Daggers' },
	{ 'Druid', 'Weapon', 'Fist Weapons' },
	{ 'Druid', 'Weapon', 'One-Handed Maces' },
	{ 'Druid', 'Weapon', 'Polearms' },
	{ 'Druid', 'Weapon', 'Staves' },
	{ 'Druid', 'Weapon', 'Two-Handed Maces' },

--	{ 'Hunter', 'Armor', 'Cloth' },
--	{ 'Hunter', 'Armor', 'Leather' },
	{ 'Hunter', 'Armor', 'Mail' },
	{ 'Hunter', 'Armor', 'Miscellaneous' },
	{ 'Hunter', 'Weapon', 'One-Handed Axes' },
	{ 'Hunter', 'Weapon', 'Daggers' },
	{ 'Hunter', 'Weapon', 'Fist Weapons' },
	{ 'Hunter', 'Weapon', 'Polearms' },
	{ 'Hunter', 'Weapon', 'Staves' },
	{ 'Hunter', 'Weapon', 'One-Handed Swords' },
	{ 'Hunter', 'Weapon', 'Two-Handed Axes' },
	{ 'Hunter', 'Weapon', 'Two-Handed Swords' },
	{ 'Hunter', 'Weapon', 'Bows' },
	{ 'Hunter', 'Weapon', 'Crossbows' },
	{ 'Hunter', 'Weapon', 'Guns' },

	{ 'Mage', 'Armor', 'Cloth' },
	{ 'Mage', 'Armor', 'Miscellaneous' },
	{ 'Mage', 'Weapon', 'Daggers' },
	{ 'Mage', 'Weapon', 'Staves' },
	{ 'Mage', 'Weapon', 'One-Handed Swords' },
	{ 'Mage', 'Weapon', 'Wands' },

--	{ 'Monk', 'Armor', 'Cloth' },
	{ 'Monk', 'Armor', 'Leather' },
	{ 'Monk', 'Armor', 'Miscellaneous' },
	{ 'Monk', 'Weapon', 'One-Handed Axes' },
	{ 'Monk', 'Weapon', 'Fist Weapons' },
	{ 'Monk', 'Weapon', 'One-Handed Maces' },
	{ 'Monk', 'Weapon', 'Polearms' },
	{ 'Monk', 'Weapon', 'Staves' },
	{ 'Monk', 'Weapon', 'One-Handed Swords' },

--	{ 'Paladin', 'Armor', 'Cloth' },
--	{ 'Paladin', 'Armor', 'Leather' },
--	{ 'Paladin', 'Armor', 'Mail' },
	{ 'Paladin', 'Armor', 'Plate' },
	{ 'Paladin', 'Armor', 'Miscellaneous' },
	{ 'Paladin', 'Armor', 'Shields' },
	{ 'Paladin', 'Weapon', 'One-Handed Axes' },
	{ 'Paladin', 'Weapon', 'One-Handed Maces' },
	{ 'Paladin', 'Weapon', 'Polearms' },
	{ 'Paladin', 'Weapon', 'One-Handed Swords' },
	{ 'Paladin', 'Weapon', 'Two-Handed Axes' },
	{ 'Paladin', 'Weapon', 'Two-Handed Maces' },
	{ 'Paladin', 'Weapon', 'Two-Handed Swords' },

	{ 'Priest', 'Armor', 'Cloth' },
	{ 'Priest', 'Armor', 'Miscellaneous' },
	{ 'Priest', 'Weapon', 'Daggers' },
	{ 'Priest', 'Weapon', 'One-Handed Maces' },
	{ 'Priest', 'Weapon', 'Staves' },
	{ 'Priest', 'Weapon', 'Wands' },

--	{ 'Rogue', 'Armor', 'Cloth' },
	{ 'Rogue', 'Armor', 'Leather' },
	{ 'Rogue', 'Armor', 'Miscellaneous' },
	{ 'Rogue', 'Weapon', 'One-Handed Axes' },
	{ 'Rogue', 'Weapon', 'Daggers' },
	{ 'Rogue', 'Weapon', 'Fist Weapons' },
	{ 'Rogue', 'Weapon', 'One-Handed Maces' },
	{ 'Rogue', 'Weapon', 'One-Handed Swords' },
	{ 'Rogue', 'Weapon', 'Bows' },
	{ 'Rogue', 'Weapon', 'Crossbows' },
	{ 'Rogue', 'Weapon', 'Guns' },

--	{ 'Shaman', 'Armor', 'Cloth' },
--	{ 'Shaman', 'Armor', 'Leather' },
	{ 'Shaman', 'Armor', 'Mail' },
	{ 'Shaman', 'Armor', 'Miscellaneous' },
	{ 'Shaman', 'Armor', 'Shields' },
	{ 'Shaman', 'Weapon', 'One-Handed Axes' },
	{ 'Shaman', 'Weapon', 'Daggers' },
	{ 'Shaman', 'Weapon', 'Fist Weapons' },
	{ 'Shaman', 'Weapon', 'One-Handed Maces' },
	{ 'Shaman', 'Weapon', 'Staves' },
	{ 'Shaman', 'Weapon', 'Two-Handed Axes' },
	{ 'Shaman', 'Weapon', 'Two-Handed Maces' },

	{ 'Warlock', 'Armor', 'Cloth' },
	{ 'Warlock', 'Armor', 'Miscellaneous' },
	{ 'Warlock', 'Weapon', 'Daggers' },
	{ 'Warlock', 'Weapon', 'One-Handed Swords' },
	{ 'Warlock', 'Weapon', 'Staves' },
	{ 'Warlock', 'Weapon', 'Wands' },

--	{ 'Warrior', 'Armor', 'Cloth' },
--	{ 'Warrior', 'Armor', 'Leather' },
--	{ 'Warrior', 'Armor', 'Mail' },
	{ 'Warrior', 'Armor', 'Plate' },
	{ 'Warrior', 'Armor', 'Miscellaneous' },
	{ 'Warrior', 'Armor', 'Shields' },
	{ 'Warrior', 'Weapon', 'One-Handed Axes' },
	{ 'Warrior', 'Weapon', 'Daggers' },
	{ 'Warrior', 'Weapon', 'Fist Weapons' },
	{ 'Warrior', 'Weapon', 'One-Handed Maces' },
	{ 'Warrior', 'Weapon', 'Polearms' },
	{ 'Warrior', 'Weapon', 'Staves' },
	{ 'Warrior', 'Weapon', 'One-Handed Swords' },
	{ 'Warrior', 'Weapon', 'Two-Handed Axes' },
	{ 'Warrior', 'Weapon', 'Two-Handed Maces' },
	{ 'Warrior', 'Weapon', 'Two-Handed Swords' },
	{ 'Warrior', 'Weapon', 'Bows' },
	{ 'Warrior', 'Weapon', 'Crossbows' },
	{ 'Warrior', 'Weapon', 'Guns' }
}

local ValidRelics = {
	[250] = {'Blood', 'Shadow', 'Iron'}, -- Blood DK
	[251] = {'Frost', 'Shadow', 'Frost'}, -- Frost DK
	[252] = {'Fire', 'Shadow', 'Blood'}, -- Unholy DK

	[577] = {'Fel', 'Shadow', 'Fel'}, -- Havoc DH
	[581] = {'Iron', 'Arcane', 'Fel'}, -- Vengeance DH

	[102] = {'Arcane', 'Life', 'Arcane'}, -- Balance Druid
	[103] = {'Frost', 'Blood', 'Life'}, -- Feral Druid
	[104] = {'Fire', 'Blood', 'Life'}, -- Guardian Druid
	[105] = {'Life', 'Frost', 'Life'}, -- Restoration Druid

	[253] = {'Wind', 'Arcane', 'Iron'}, -- Beast Mastery Hunter
	[254] = {'Wind', 'Blood', 'Life'}, -- Marksmanship Hunter
	[255] = {'Wind', 'Iron', 'Blood'}, -- Survival Hunter

	[62] = {'Arcane', 'Frost', 'Arcane'}, -- Arcane Mage
	[63] = {'Fire', 'Arcane', 'Fire'}, -- Fire Mage
	[64] = {'Frost', 'Arcane', 'Frost'}, -- Frost Mage

	[268] = {'Life', 'Wind', 'Iron'}, -- Brewmaster Monk
	[270] = {'Frost', 'Life', 'Wind'}, -- Mistweaver Monk
	[269] = {'Wind', 'Iron', 'Wind'}, -- Windwalker Monk

	[65] = {'Holy', 'Life', 'Holy'}, -- Holy Paladin
	[66] = {'Holy', 'Iron', 'Arcane'}, -- Protection Paladin
	[70] = {'Holy', 'Fire', 'Holy'}, -- Retribution Paladin

	[256] = {'Holy', 'Shadow', 'Holy'}, -- Discipline Priest
	[257] = {'Holy', 'Life', 'Holy'}, -- Holy Priest
	[258] = {'Shadow', 'Blood', 'Shadow'}, -- Shadow Priest

	[259] = {'Shadow', 'Iron', 'Blood'}, -- Assassination Rogue
	[260] = {'Blood', 'Iron', 'Wind'}, -- Outlaw Rogue
	[261] = {'Fel', 'Shadow', 'Fel'}, -- Subtlety Rogue

	[262] = {'Wind', 'Frost', 'Wind'}, -- Elemental Shaman
	[263] = {'Fire', 'Iron', 'Wind'}, -- Enhancement Shaman
	[264] = {'Life', 'Frost', 'Life'}, -- Restoration Shaman

	[265] = {'Shadow', 'Blood', 'Shadow'}, -- Affliction Warlock
	[266] = {'Shadow', 'Fire', 'Fel'}, -- Demonology Warlock
	[267] = {'Fel', 'Fire', 'Fel'}, -- Destruction Warlock

	[71] = {'Iron', 'Blood', 'Shadow'}, -- Arms Warrior
	[72] = {'Fire', 'Wind', 'Iron'}, -- Fury Warrior
	[73] = {'Iron', 'Blood', 'Fire'}, -- Protection Warrior
}

-- IDs for following are from http://wow.gamepedia.com/API_GetInspectSpecialization
local PrimaryAttributes = {
	{ 'Death Knight', 'Any', 'Strength' },
	
	{ 'Demon Hunter', 'Any', 'Agility' },
	
	{ 'Druid', 102, 'Intellect' },			-- balance
	{ 'Druid', 103, 'Agility' },			-- feral
	{ 'Druid', 104, 'Agility' },			-- guardian
	{ 'Druid', 105, 'Intellect' },			-- restoration
	
	{ 'Hunter', 'Any', 'Agility' },
	
	{ 'Mage', 'Any', 'Intellect' },

	{ 'Monk', 268, 'Agility' },				-- brewmaster
	{ 'Monk', 270, 'Intellect' },			-- mistweaver
	{ 'Monk', 269, 'Agility' },				-- windwalker
	
	{ 'Paladin', 65, 'Intellect' },			-- holy
	{ 'Paladin', 66, 'Strength' },			-- protection
	{ 'Paladin', 70, 'Strength' },			-- retribution

	{ 'Priest', 'Any', 'Intellect' },
	
	{ 'Rogue', 'Any', 'Agility' },

	{ 'Shaman', 262, 'Intellect' },			-- elemental
	{ 'Shaman', 263, 'Agility' },			-- enhancement
	{ 'Shaman', 264, 'Intellect' },			-- restoration
	
	{ 'Warlock', 'Any', 'Intellect' },
	
	{ 'Warrior', 'Any', 'Strength' }
}
	
local OffspecAttributes = {
	{ 'Druid', 102, 'Agility' },			-- balance
	{ 'Druid', 103, 'Intellect' },			-- feral
	{ 'Druid', 104, 'Intellect' },			-- guardian
	{ 'Druid', 105, 'Agility' },			-- restoration

	{ 'Monk', 268, 'Intellect' },			-- brewmaster
	{ 'Monk', 270, 'Agility' },				-- mistweaver
	{ 'Monk', 269, 'Intellect' },			-- windwalker

	{ 'Paladin', 65, 'Strength' },			-- holy
	{ 'Paladin', 66, 'Intellect' },			-- protection
	{ 'Paladin', 70, 'Intellect' },			-- retribution

	{ 'Shaman', 262, 'Agility' },			-- elemental
	{ 'Shaman', 263, 'Intellect' },			-- enhancement
	{ 'Shaman', 264, 'Agility' }			-- restoration
	
}

-- note that the following is only valid for Legion; previously different weapons were possible (ex: feral staff vs. 2 fist weapons)
-- adding 3 for relics
local ExpectedItemCount = {			-- number of items expected by spec, assuming person has gear in every slot
	{ 'Death Knight', 250, 15 },			-- blood
	{ 'Death Knight', 251, 16 },			-- frost
	{ 'Death Knight', 252, 15 },			-- unholy
	
	{ 'Demon Hunter', 577, 16 },			-- havoc
	{ 'Demon Hunter', 581, 16 },			-- vengeance
	
	{ 'Druid', 102, 15 },					-- balance
	{ 'Druid', 103, 16 },					-- feral
	{ 'Druid', 104, 16 },					-- guardian
	{ 'Druid', 105, 15 },					-- restoration
	
	{ 'Hunter', 253, 15 },					-- beast mastery
	{ 'Hunter', 254, 15 },					-- marksmanship
	{ 'Hunter', 255, 15 },					-- survival
	
	{ 'Mage', 62, 15 },						-- arcane
	{ 'Mage', 63, 16 },						-- fire
	{ 'Mage', 64, 15 },						-- frost

	{ 'Monk', 268, 15 },					-- brewmaster
	{ 'Monk', 270, 15 },					-- mistweaver
	{ 'Monk', 269, 16 },					-- windwalker
	
	{ 'Paladin', 65, 15 },					-- holy
	{ 'Paladin', 66, 16 },					-- protection
	{ 'Paladin', 70, 15 },					-- retribution

	{ 'Priest', 256, 15 },					-- discipline
	{ 'Priest', 257, 15 },					-- holy
	{ 'Priest', 258, 16 },					-- shadow
	
	{ 'Rogue', 259, 16 },					-- assassination
	{ 'Rogue', 260, 16 },					-- outlaw
	{ 'Rogue', 261, 16 },					-- subtlety

	{ 'Shaman', 262, 16 },					-- elemental
	{ 'Shaman', 263, 16 },					-- enhancement
	{ 'Shaman', 264, 16 },					-- restoration
	
	{ 'Warlock', 256, 15 },					-- affliction
	{ 'Warlock', 266, 16 },					-- demonology
	{ 'Warlock', 267, 15 },					-- destruction
	
	{ 'Warrior', 71, 15 },					-- arms
	{ 'Warrior', 72, 16 },					-- fury
	{ 'Warrior', 73, 16 }					-- protection
}

--[[
to easily populate these arrays:
	ex: http://www.wowhead.com/items/armor/trinkets/role:1?filter=166;7;0
	wowhead search trinkets -> usable by "whichever" -> added in expansion/patch
	paste into OpenOffice
	" in column b
	", in column d
	=concatenate(b2;c2;d2) in column e
	ensure curly quotes are off in tools -> autocorrect options -> localized options
]]--	
local TRINKET_AGILITY_DPS = {
	"Arcanogolem Digit",
	"Entwined Elemental Foci",
	"Fury of the Burning Sky",
	"Nightblooming Frond",
	"Convergence of Fates",
	"Draught of Souls",
	"Icon of Rot",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Conquest",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Conquest",
	"Ravaged Seed Pod",
	"Twisting Wind",
	"Spontaneous Appendages",
	"Bloodthirsty Instinct",
	"Nature's Call",
	"Unstable Horrorslime",
	"Legion Season 1Vindictive Gladiator's Insignia of Conquest",
	"Legion Season 1Vindictive Gladiator's Insignia of Conquest",
	"Eye of Skovald",
	"Memento of Angerboda",
	"Spiked Counterweight",
	"Caged Horror",
	"Hunger of the Pack",
	"Oakheart's Gnarled Root",
	"Nightmare Egg Shell",
	"Figurehead of the Naglfar",
	"Naraxas' Spiked Tongue",
	"Mark of Dargrul",
	"Stormsinger Fulmination Charge",
	"Giant Ornamental Pearl",
	"Tempered Egg of Serpentrix",
	"Terrorbound Nexus",
	"Obelisk of the Void",
	"Tiny Oozeling in a Jar",
	"Elementium Bomb Squirrel Generator",
	"Chaos Talisman",
	"Windscar Whetstone",
	"Tirathon's Betrayal",
	"Faulty Countermeasure",
	"Moonlit Prism",
	"The Devilsaur's Bite",
	"Legion Season 1Vindictive Combatant's Insignia of Conquest",
	"Legion Season 1Vindictive Combatant's Insignia of Conquest",
	"Etching of SargerasDemon Hunter",
	"Lekos' LeashDemon Hunter",
	"Frothing Helhound's Fury",
	"Spine of Barax"
}

local TRINKET_INTELLECT_DPS = {
	"Erratic Metronome",
	"Icon of Rot",
	"Fury of the Burning Sky",
	"Star Gate",
	"Whispers in the Dark",
	"Pharamere's Forbidden Grimore",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Dominance",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Dominance",
	"Swarming Plaguehive",
	"Twisting Wind",
	"Wriggling Sinew",
	"Bough of Corruption",
	"Unstable Horrorslime",
	"Legion Season 1Vindictive Gladiator's Insignia of Dominance",
	"Legion Season 1Vindictive Gladiator's Insignia of Dominance",
	"Eye of Skovald",
	"Caged Horror",
	"Corrupted Starlight",
	"Oakheart's Gnarled Root",
	"Figurehead of the Naglfar",
	"Naraxas' Spiked Tongue",
	"Stormsinger Fulmination Charge",
	"Portable Manacracker",
	"Obelisk of the Void",
	"Elementium Bomb Squirrel Generator",
	"Infernal Writ",
	"Moonlit Prism",
	"Devilsaur Shock-Baton",
	"Legion Season 1Vindictive Combatant's Insignia of Dominance",
	"Legion Season 1Vindictive Combatant's Insignia of Dominance",
	"Pocket Void Portal",
	"Ancient Leaf",
	"Runas' Nearly Depleted Ley Crystal",
	"The Watcher's Divine Inspiration"
}

local TRINKET_STRENGTH_DPS = {
	"Claw of the Crystalline Scorpid",
	"Entwined Elemental Foci",
	"Convergence of Fates",
	"Draught of Souls",
	"Might of Krosus",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Victory",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Victory",
	"Spontaneous Appendages",
	"Ursoc's Rending Paw",
	"Nature's Call",
	"Ravaged Seed Pod",
	"Legion Season 1Vindictive Gladiator's Insignia of Victory",
	"Legion Season 1Vindictive Gladiator's Insignia of Victory",
	"Memento of Angerboda",
	"Spiked Counterweight",
	"Hunger of the Pack",
	"Nightmare Egg Shell",
	"Mark of Dargrul",
	"Giant Ornamental Pearl",
	"Terrorbound Nexus",
	"Tiny Oozeling in a Jar",
	"Chaos Talisman",
	"Windscar Whetstone",
	"Faulty Countermeasure",
	"Impact Tremor",
	"Legion Season 1Vindictive Combatant's Insignia of Victory",
	"Legion Season 1Vindictive Combatant's Insignia of Victory",
	"Mountain Rage Shaker",
	"Might of the Forsaken",
	"Tenacity of Cursed Blood"
}

local TRINKET_HEALER = {
	"Perfectly Preserved Cake",
	"Aluriel's Mirror",
	"Ephemeral Paradox",
	"Etraeus' Celestial Map",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Dominance",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Dominance",
	"Vial of Nightmare Fog",
	"Heightened Senses",
	"Horn of Cenarius",
	"Cocoon of Enforced Solitude",
	"Legion Season 1Vindictive Gladiator's Insignia of Dominance",
	"Legion Season 1Vindictive Gladiator's Insignia of Dominance",
	"Eye of Skovald",
	"Naglfar Fare",
	"Mote of Sanctification",
	"Nether Anti-Toxin",
	"Amalgam's Seventh Spine",
	"Caged Horror",
	"Corrupted Starlight",
	"Oakheart's Gnarled Root",
	"Figurehead of the Naglfar",
	"Naraxas' Spiked Tongue",
	"Stormsinger Fulmination Charge",
	"Bottled Hurricane",
	"Portable Manacracker",
	"Obelisk of the Void",
	"Elementium Bomb Squirrel Generator",
	"Thrumming Gossamer",
	"Jewel of Insatiable Desire",
	"Flask of the Solemn Night",
	"Infernal Writ",
	"Concave Reflecting Lens",
	"Moonlit Prism",
	"Devilsaur Shock-Baton",
	"Legion Season 1Vindictive Combatant's Insignia of Dominance",
	"Legion Season 1Vindictive Combatant's Insignia of Dominance",
	"Pocket Void Portal",
	"Ancient Leaf",
	"Runas' Nearly Depleted Ley Crystal",
	"The Watcher's Divine Inspiration",
	"Mana-Infused Berries"		-- TEST
}

local TRINKET_TANK = {
	"Animated Exoskeleton",
	"Royal Dagger Haft",
	"Infernal Contract",
	"Fang of Tichcondrius",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Victory",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Victory",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Conquest",
	"Legion Season 1 EliteVindictive Gladiator's Insignia of Conquest",
	"Goblet of Nightmarish Ichor",
	"Unbridled Fury",
	"Phantasmal Echo",
	"Grotesque Statuette",
	"Legion Season 1Vindictive Gladiator's Insignia of Conquest",
	"Legion Season 1Vindictive Gladiator's Insignia of Victory",
	"Legion Season 1Vindictive Gladiator's Insignia of Conquest",
	"Legion Season 1Vindictive Gladiator's Insignia of Victory",
	"Gift of Radiance",
	"Ember of Nullification",
	"Writhing Heart of Darkness",
	"Shard of Rokmora",
	"Talisman of the Cragshaper",
	"Parjesh's Medallion",
	"Coagulated Nightwell Residue",
	"Impenetrable Nerubian Husk",
	"Shivermaw's Jawbone",
	"Orb of Torment",
	"The Devilsaur's Bite",
	"Impact Tremor",
	"Legion Season 1Vindictive Combatant's Insignia of Conquest",
	"Legion Season 1Vindictive Combatant's Insignia of Conquest",
	"Legion Season 1Vindictive Combatant's Insignia of Victory",
	"Legion Season 1Vindictive Combatant's Insignia of Victory",
	"Mountain Rage Shaker",
	"Etching of SargerasDemon Hunter",
	"Might of the Forsaken",
	"Lekos' LeashDemon Hunter",
	"Tenacity of Cursed Blood",
	"Frothing Helhound's Fury",
	"Spine of Barax"
}

-- set up Frames that will listen for events
local lootReceivedEventFrame
local whisperReceivedEventFrame
local rollReceivedEventFrame
local inspectReadyEventFrame
local rosterUpdatedEventFrame
local combatStatusChangedEventFrame
local rollDelayFrame
local groupMemberInfoChangedEventFrame

-- set up variables that will track addon's status
local isEnabled = false
local delay = 0
local nextRollDelay = 0
local priorCacheRefreshTime = 0

local whisperedItems = {}  -- list of items we've whispered to people; index is character name-realm, content is item

local numOfQueuedRollItems = 0;
local queuedRollOwners = {}
local queuedRollItems = {}

local currentRollOwner = nil
local currentRollItem = nil
local currentRolls = {}

-- set up variables that will cache group member's information
local groupInfoCache = {}  -- array keyed by name-realm paired with a list of slotid-items
local maxInspectIndex = 0  -- the index of the last character in GetRaidRosterInfo(); must be < inspectIndex to start with so PopulateGroupInfoCache can start inspections
local inspectIndex = maxInspectIndex + 1    -- the index of the character within GetRaidRosterInfo() that we're currently inspecting
											-- defaulting to a higher value so PopulateGroupCache() knows to start a new inspection loop the first time it is called
local notifyInspectName = nil -- valued if we sent a request to inspect someone, nil otherwise
--local inspectRetries = 0  -- tracks how many times we've attempted to inspect a specific character without getting an INSPECT_READY result
local inspectLoop = MAX_INSPECT_LOOPS + 1 -- defaulting to a higher value so PopulateGroupCache() knows to start a new inspection loop the first time it is called

local function GetExpectedRelicCount(level)
	if level ~= nil then
		if level >= 101 then
			return NUM_EXPECTED_RELICS
		end
	end
	return 0
end

local function GetExpectedItemCount(class, spec, level)
	if class ~= nil and spec ~= nil and level ~= nil then
		if level >= 101 then   -- pre-Legion, classes could use different weapons; ex: feral used a staff instead of 2 fist weapons
								-- technically the change happens once the person gets their artifact, but we'll just simplify and
								-- expect Legion counts at 101 and above, and 15 items for 100 and earlier
			local i = 1
			while ExpectedItemCount[i] do
				if class == ExpectedItemCount[i][1] and spec == ExpectedItemCount[i][2] then
					return ExpectedItemCount[i][3]
				end
				i = i + 1
			end
		end
	end
	return NUM_EXPECTED_ITEMS
end

function PLH_GetRelicCountFromCache(name)
	local relicCount = 0
	local characterDetails = groupInfoCache[name]
	if characterDetails ~= nil then
		if characterDetails[PLH_RELICSLOT + 1] ~= nil then
			relicCount = relicCount + 1
		end
		if characterDetails[PLH_RELICSLOT + 2] ~= nil then
			relicCount = relicCount + 1
		end
		if characterDetails[PLH_RELICSLOT + 3] ~= nil then
			relicCount = relicCount + 1
		end
	end
	return relicCount
end

-- only returns count of equippable items from cache; excludes relics and other cached info such as ClassName/Spec/Level
function PLH_GetItemCountFromCache(name)
	local itemCount = 0
	if groupInfoCache[name] ~= nil then
		for slotID, item in pairs (groupInfoCache[name]) do
			itemCount = itemCount + 1
		end
		itemCount = itemCount - 3 -- subtract 3 since everyone has ClassName and Spec and Level elements
		itemCount = itemCount - PLH_GetRelicCountFromCache(name)
	else
		PLH_SendDebugMessage('groupInfoCache[name] is nil for ', name)	
	end
	return itemCount
end

-- Returns the item that character has equipped in slotID, based on the cache
local function GetEquippedItem(characterName, slotID)
	local item = nil
	if characterName == PLH_GetUnitNameWithRealm('player') then
		-- we can get the item directly
		item = GetInventoryItemLink('player', slotID)
	else
		-- we have to get the item from the cache
		local characterDetails = groupInfoCache[characterName]
		if characterDetails ~= nil then
			item = characterDetails[slotID]
		end
	end
	return item
end

local function GetEquippedRelic(characterName, relicNumber)
	local relic = nil
	if characterName == PLH_GetUnitNameWithRealm('player') then
		-- we can get the item directly
		local weapon = GetInventoryItemLink('player', INVSLOT_MAINHAND)
		_, relic = GetItemGem(weapon, relicNumber)
	else
		-- we have to get the item from the cache
		local characterDetails = groupInfoCache[characterName]
		if characterDetails ~= nil then
			relic = characterDetails[PLH_RELICSLOT + relicNumber]
		end
	end
	return relic
end

-- note that this will return a value based on player's class/spec, so it will switch if the primary attribute is mutable!
-- thus only use for cloaks/rings/necks/trinkets, sine those things are not mutable
local function GetItemPrimaryAttribute(item)
	local stats = GetItemStats(item)
	if stats ~= nil then
		for stat, value in pairs(stats) do
			if _G[stat] == 'Strength' or _G[stat] == 'Intellect' or _G[stat] == 'Agility' then
				return _G[stat]
			end
		end
	end
	return nil
end

-- TODO may need to limit this by ilvl?  ex: did cloaks only become mutable in Legion?
local function IsMutablePrimaryAttribute(itemEquipLoc)
	return itemEquipLoc == 'INVTYPE_HEAD'
		or itemEquipLoc == 'INVTYPE_SHOULDER'
		or itemEquipLoc == 'INVTYPE_CLOAK'
		or itemEquipLoc == 'INVTYPE_CHEST'
		or itemEquipLoc == 'INVTYPE_ROBE'
		or itemEquipLoc == 'INVTYPE_WAIST'
		or itemEquipLoc == 'INVTYPE_LEGS'
		or itemEquipLoc == 'INVTYPE_FEET'
		or itemEquipLoc == 'INVTYPE_WRIST'
		or itemEquipLoc == 'INVTYPE_HAND'
end

local function IsTrinketUsable(item, role)
	local itemName = select(1, GetItemInfo(item))
	local trinketList = nil
	if role == 'AgilityDPS' then
		trinketList = TRINKET_AGILITY_DPS
	elseif role == 'IntellectDPS' then
		trinketList = TRINKET_INTELLECT_DPS
	elseif role == 'StrengthDPS' then
		trinketList = TRINKET_STRENGTH_DPS
	elseif role == 'Healer' then
		trinketList = TRINKET_HEALER
	elseif role == 'Tank' then
		trinketList = TRINKET_TANK
	end
	
	local i
	if trinketList ~= nil then
		for i = 1, #trinketList do
			if trinketList[i] == itemName then
				return true
			end
		end
	end
	
	return false
end

local function IsRelic(item)
	local _, _, _, _, _, itemType, itemSubtype, _, _, _, _ = GetItemInfo(item)
	return itemType == 'Gem' and itemSubtype == 'Artifact Relic'
end

local function IsValidRelicTypeForSpec(relicType, spec)
	local specRelics = ValidRelics[spec]
	if specRelics ~= nil then
		return ValidRelics[spec][1] == relicType or ValidRelics[spec][2] == relicType or ValidRelics[spec][3] == relicType
	else
		return false
	end
end

-- Returns false if the character cannot use the item.
local function IsEquippableItemForCharacter(item, characterName)
	local isEquippableForClass = false
	local isEquippableForSpec = false
	local isEquippableForOffspec = false
	if IsEquippableItem(item) or IsRelic(item) then
		local _, _, _, _, _, itemType, itemSubtype, _, itemEquipLoc, _, _ = GetItemInfo(item)
		local class
		local spec
		if characterName == PLH_GetUnitNameWithRealm('player') then
			class = select(1, GetPlayerInfoByGUID(PLH_GetUnitGUIDFromFullname(characterName)))
			spec = GetSpecializationInfo(GetSpecialization())
		elseif groupInfoCache[characterName] ~= nil then
			class = groupInfoCache[characterName]['ClassName']
			spec = groupInfoCache[characterName]['Spec']
		else
			PLH_SendDebugMessage('Unable to determine class and spec in InEquippableItemForCharacter()!!!!')
			return false  -- should never reach here, but if we do it means we're not looking up the player or anyone in cache
		end
		local isRelic = IsRelic(item)
		isEquippableForClass = itemEquipLoc == 'INVTYPE_CLOAK' -- cloaks show up as type=armor, subtype=cloth, but they're equippable by all, so set to true if cloak
		local i = 1
		while not isEquippableForClass and ValidGear[i] do
			if class == ValidGear[i][1] and itemType == ValidGear[i][2] and itemSubtype == ValidGear[i][3] then
				isEquippableForClass = true
			end
			i = i + 1
		end

		if isEquippableForClass then
			if itemEquipLoc == 'INVTYPE_TRINKET' then
				if spec == 105 or spec == 270 or spec == 65 or spec == 256 or spec == 257 or spec == 264 then
					isEquippableForSpec = IsTrinketUsable(item, 'Healer')
				elseif spec == 250 or spec == 581 or spec == 104 or spec == 268 or spec == 66 or spec == 73 then
					isEquippableForSpec = IsTrinketUsable(item, 'Tank')
				elseif spec == 577 or spec == 103 or spec == 253 or spec == 254 or spec == 255 or spec == 269 or spec == 259 or spec == 260 or spec == 261 or spec == 263 then
					isEquippableForSpec = IsTrinketUsable(item, 'AgilityDPS')
				elseif spec == 251 or spec == 252 or spec == 70 or spec == 71 or spec == 72 then
					isEquippableForSpec = IsTrinketUsable(item, 'StrengthDPS')
				elseif spec == 102 or spec == 62 or spec == 63 or spec == 64 or spec == 258 or spec == 262 or spec == 265 or spec == 266 or spec == 267 then
					isEquippableForSpec = IsTrinketUsable(item, 'IntellectDPS')
				end
					
				if not PLH_CURRENT_SPEC_ONLY and not isEquippableForSpec then
					if class == 'Death Knight' or class == 'Warrior' then
						isEquippableForOffspec = IsTrinketUsable(item, 'Tank') or IsTrinketUsable(item, 'StrengthDPS')
					elseif class == 'Demon Hunter' then
						isEquippableForOffspec = IsTrinketUsable(item, 'Tank') or IsTrinketUsable(item, 'AgilityDPS')
					elseif class == 'Druid' then
						isEquippableForOffspec = IsTrinketUsable(item, 'Tank') or IsTrinketUsable(item, 'AgilityDPS') or IsTrinketUsable(item, 'Healer') or IsTrinketUsable(item, 'IntellectDPS')
					elseif class == 'Monk' then
						isEquippableForOffspec = IsTrinketUsable(item, 'Tank') or IsTrinketUsable(item, 'AgilityDPS') or IsTrinketUsable(item, 'Healer')
					elseif class == 'Paladin' then
						isEquippableForOffspec = IsTrinketUsable(item, 'Tank') or IsTrinketUsable(item, 'StrengthDPS') or IsTrinketUsable(item, 'Healer')
					elseif class == 'Priest' then
						isEquippableForOffspec = IsTrinketUsable(item, 'Healer') or IsTrinketUsable(item, 'IntellectDPS')
					elseif class == 'Shaman' then
						isEquippableForOffspec = IsTrinketUsable(item, 'Healer') or IsTrinketUsable(item, 'IntellectDPS') or IsTrinketUsable(item, 'AgilityDPS')
					end
				end
			else
				itemPrimaryAttribute = GetItemPrimaryAttribute(item)
				if itemPrimaryAttribute == nil then
					isEquippableForSpec = true  -- if there's no primary attr (ex: ring/neck), then the item is equippable by everyone
				elseif IsMutablePrimaryAttribute(itemEquipLoc) then
					isEquippableForSpec = true	-- if the item is a piece of gear that has mutable primary stats then return true
				else
					-- otherwise we're going to check if the item's primary attribute is applicable for the character's spec
					i = 1
					while not isEquippableForSpec and PrimaryAttributes[i] do
						if class == PrimaryAttributes[i][1] and PrimaryAttributes[i][3] == itemPrimaryAttribute and (PrimaryAttributes[i][2] == 'Any' or PrimaryAttributes[i][2] == spec) then
							isEquippableForSpec = true
						end
						i = i + 1
					end

					if not PLH_CURRENT_SPEC_ONLY and not isEquippableForSpec then
						-- now check to see if it's usable by an offspec
						i = 1
						while not isEquippableForSpec and OffspecAttributes[i] do
							if class == OffspecAttributes[i][1] and OffspecAttributes[i][3] == itemPrimaryAttribute and OffspecAttributes[i][2] == spec then
								isEquippableForOffspec = true
							end
							i = i + 1
						end
					end
				end
			end
		elseif isRelic then
			local relicType = PLH_GetRelicType(item)
			isEquippableForSpec = IsValidRelicTypeForSpec(relicType, spec)
			isEquippableForClass = isEquippableForSpec
--[[  No point in checking offspec since we can't inspect the off-spec Artifact to see what relics it has!
			if not PLH_CURRENT_SPEC_ONLY and not isEquippableForSpec then
				if class == 'Death Knight' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 250) or IsValidRelicTypeForSpec(relicType, 251) or IsValidRelicTypeForSpec(relicType, 252)
				elseif class == 'Demon Hunter' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 577) or IsValidRelicTypeForSpec(relicType, 581)
				elseif class == 'Druid' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 102) or IsValidRelicTypeForSpec(relicType, 103) or IsValidRelicTypeForSpec(relicType, 104) or IsValidRelicTypeForSpec(relicType, 105)
				elseif class == 'Hunter' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 253) or IsValidRelicTypeForSpec(relicType, 254) or IsValidRelicTypeForSpec(relicType, 255)
				elseif class == 'Mage' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 62) or IsValidRelicTypeForSpec(relicType, 63) or IsValidRelicTypeForSpec(relicType, 64)
				elseif class == 'Monk' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 268) or IsValidRelicTypeForSpec(relicType, 270) or IsValidRelicTypeForSpec(relicType, 269)
				elseif class == 'Paladin' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 65) or IsValidRelicTypeForSpec(relicType, 66) or IsValidRelicTypeForSpec(relicType, 70)
				elseif class == 'Priest' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 256) or IsValidRelicTypeForSpec(relicType, 257) or IsValidRelicTypeForSpec(relicType, 258)
				elseif class == 'Rogue' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 259) or IsValidRelicTypeForSpec(relicType, 260) or IsValidRelicTypeForSpec(relicType, 261)
				elseif class == 'Shaman' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 262) or IsValidRelicTypeForSpec(relicType, 263) or IsValidRelicTypeForSpec(relicType, 264)
				elseif class == 'Warlock' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 265) or IsValidRelicTypeForSpec(relicType, 266) or IsValidRelicTypeForSpec(relicType, 267)
				elseif class == 'Warrior' then
					isEquippableForOffspec = IsValidRelicTypeForSpec(relicType, 71) or IsValidRelicTypeForSpec(relicType, 72) or IsValidRelicTypeForSpec(relicType, 73)
				end
				isEquippableForClass = isEquippableForSpec or isEquippableForOffspec
			end
]]--			
		end
	end

--	PLH_SendDebugMessage('For ' .. characterName .. ' item ' .. item .. ' isEquippableForClass = ' .. tostring(isEquippableForClass) .. ' and isEquippableForSpec = ' .. tostring(isEquippableForSpec)  .. ' and isEquippableForOffspec = ' .. tostring(isEquippableForOffspec))
	
	return isEquippableForClass and (isEquippableForSpec or isEquippableForOffspec)
end

-- returns two variables:  true if the item is an upgrade over equippedItem (based on ilvl), equipped ilvl
local function IsAnUpgrade(item, equippedItem)
	local equippedILVL = PLH_GetRealILVL(equippedItem)
	if equippedILVL == 0 then
		-- this means we couldn't find an equippedItem
		return false, 0
	else
		return PLH_GetRealILVL(item) > PLH_GetRealILVL(equippedItem), equippedILVL
	end
end

-- returns an appropriate SlotID for the given itemEquipLoc, or nil if it's not an item
-- if itemEquipLoc is a finger slot or trinket slot, we'll just return the first item
-- if itemEquipLoc is a weapon that can be in either slot (INVTYPE_WEAPON), we'll return the main hand
local function GetSlotID(itemEquipLoc)
	if itemEquipLoc == 'INVTYPE_HEAD' then return INVSLOT_HEAD
	elseif itemEquipLoc == 'INVTYPE_NECK' then return INVSLOT_NECK
	elseif itemEquipLoc == 'INVTYPE_SHOULDER' then return INVSLOT_SHOULDER
	elseif itemEquipLoc == 'BODY' then return INVSLOT_BODY
	elseif itemEquipLoc == 'INVTYPE_CHEST' then return INVSLOT_CHEST
	elseif itemEquipLoc == 'INVTYPE_ROBE' then return INVSLOT_CHEST
	elseif itemEquipLoc == 'INVTYPE_WAIST' then return INVSLOT_WAIST
	elseif itemEquipLoc == 'INVTYPE_LEGS' then return INVSLOT_LEGS
	elseif itemEquipLoc == 'INVTYPE_FEET' then return INVSLOT_FEET
	elseif itemEquipLoc == 'INVTYPE_WRIST' then return INVSLOT_WRIST
	elseif itemEquipLoc == 'INVTYPE_HAND' then return INVSLOT_HAND
	elseif itemEquipLoc == 'INVTYPE_FINGER' then return INVSLOT_FINGER1
	elseif itemEquipLoc == 'INVTYPE_TRINKET' then return INVSLOT_TRINKET1
	elseif itemEquipLoc == 'INVTYPE_CLOAK' then return INVSLOT_BACK
	elseif itemEquipLoc == 'INVTYPE_WEAPON' then return INVSLOT_MAINHAND
	elseif itemEquipLoc == 'INVTYPE_SHIELD' then return INVSLOT_OFFHAND
	elseif itemEquipLoc == 'INVTYPE_2HWEAPON' then return INVSLOT_MAINHAND
	elseif itemEquipLoc == 'INVTYPE_WEAPONMAINHAND' then return INVSLOT_MAINHAND
	elseif itemEquipLoc == 'INVTYPE_WEAPONOFFHAND' then return INVSLOT_OFFHAND
	elseif itemEquipLoc == 'INVTYPE_HOLDABLE' then return INVSLOT_OFFHAND
	elseif itemEquipLoc == 'INVTYPE_RANGED' then return INVSLOT_MAINHAND
	elseif itemEquipLoc == 'INVTYPE_THROWN' then return INVSLOT_MAINHAND
	elseif itemEquipLoc == 'INVTYPE_RANGEDRIGHT' then return INVSLOT_MAINHAND
	elseif itemEquipLoc == 'INVTYPE_RELIC' then return INVSLOT_MAINHAND
	elseif itemEquipLoc == 'INVTYPE_TABARD' then return INVSLOT_TABARD
	else return nil
	end
end

-- returns two variables:  true if the item is an upgrade over equippedItem (based on ilvl), equipped ilvl
-- note: doesn't check if item is equippable, so make sure you do that check beforehand
local function IsAnUpgradeForCharacter(item, characterName)
	local isAnUpgrade = false
	local equippedILVL = 0
	local _, _, _, _, _, _, _, _, itemEquipLoc, _, _ = GetItemInfo(item)
	if itemEquipLoc ~= nil and itemEquipLoc ~= '' then
		if itemEquipLoc == 'INVTYPE_FINGER' then
			local equippedItem0 = GetEquippedItem(characterName, 11)	-- 1st ring
			local equippedItem1 = GetEquippedItem(characterName, 12)  	-- 2nd ring
			isAnUpgrade = IsAnUpgrade(item, equippedItem0) or IsAnUpgrade(item, equippedItem1)
			equippedILVL = min(PLH_GetRealILVL(equippedItem0), PLH_GetRealILVL(equippedItem1))
		elseif itemEquipLoc == 'INVTYPE_TRINKET' then
			local equippedItem0 = GetEquippedItem(characterName, 13)	-- 1st trinket
			local equippedItem1 = GetEquippedItem(characterName, 14)	-- 2nd trinket
			isAnUpgrade = IsAnUpgrade(item, equippedItem0) or IsAnUpgrade(item, equippedItem1)
			equippedILVL = min(PLH_GetRealILVL(equippedItem0), PLH_GetRealILVL(equippedItem1))
		elseif itemEquipLoc == 'INVTYPE_WEAPON' then
			local equippedItem0 = GetEquippedItem(characterName, 16)		-- main hand
			local equippedItem1 = GetEquippedItem(characterName, 17)		-- off hand
			isAnUpgrade = IsAnUpgrade(item, equippedItem0) or IsAnUpgrade(item, equippedItem1)
			equippedILVL = min(PLH_GetRealILVL(equippedItem0), PLH_GetRealILVL(equippedItem1))
		else
			local slotID = GetSlotID(itemEquipLoc)
			local equippedItem =  GetEquippedItem(characterName, slotID)
			isAnUpgrade, equippedILVL = IsAnUpgrade(item, equippedItem)
		end
	elseif IsRelic(item) then
		local relicType = PLH_GetRelicType(item)

		local relic1 = GetEquippedRelic(characterName, 1)
		local relic1ILVL = PLH_GetRealILVL(relic1)
		local relic2 = GetEquippedRelic(characterName, 2)
		local relic2ILVL = PLH_GetRealILVL(relic2)
		local relic3 = GetEquippedRelic(characterName, 3)
		local relic3ILVL = PLH_GetRealILVL(relic3)
		isAnUpgrade = (relicType == PLH_GetRelicType(relic1) and IsAnUpgrade(item, relic1))
			or (relicType == PLH_GetRelicType(relic2) and IsAnUpgrade(item, relic2))
			or (relicType == PLH_GetRelicType(relic3) and IsAnUpgrade(item, relic3))

		if relicType == PLH_GetRelicType(relic1) then
			equippedILVL = relic1ILVL
		end
		if relicType == PLH_GetRelicType(relic2) and (equippedILVL == 0 or relic2ILVL < equippedILVL) then
			equippedILVL = relic2ILVL
		end
		if relicType == PLH_GetRelicType(relic3) and (equippedILVL == 0 or relic3ILVL < equippedILVL) then
			equippedILVL = relic3ILVL
		end
	end
	return isAnUpgrade, equippedILVL
end

-- returns two variables:  first is true or false, second is list of people for whom the item may is an upgrade (by ilvl)
-- TOOD make local
 function IsAnUpgradeForAnyCharacter(item)
	local isAnUpgrade, equippedILVL
	local isAnUpgradeForAnyCharacterNames = {}

	local index = 1
	local characterName
	while GetRaidRosterInfo(index) ~= nil do
		characterName = select(1, GetRaidRosterInfo(index))
		-- this characterName may be a full name-realm or it may just be a name, but the functions we're calling expect name-realm
		if not string.find(characterName, '-') then
			characterName = PLH_GetFullName(characterName, GetRealmName())
		end
		if IsEquippableItemForCharacter(item, characterName) then
			isAnUpgrade, equippedILVL = IsAnUpgradeForCharacter(item, characterName)
			if isAnUpgrade then
--				PLH_SendDebugMessage(item .. ' is an upgrade for ' .. characterName)
				isAnUpgradeForAnyCharacterNames[#isAnUpgradeForAnyCharacterNames + 1] = PLH_GetNameWithoutRealm(characterName) .. ' (' .. equippedILVL .. ')'
			end
		end
		index = index + 1
	end

	return #isAnUpgradeForAnyCharacterNames > 0, isAnUpgradeForAnyCharacterNames
end

local function IsPlayerInUpgradeList(list)
	if list ~= nil then
		local playerName = UnitName('player')
		for i = 1, #list do
			if string.sub(list[i], 1, string.len(playerName)) == playerName then  -- see if string starts with 'playername '; ex: 'Madone (690)'
				return true
			end
		end
	end
	return false
end

-- returns the names from the given array, with 'and others' if array size > limit
local function GetNames(namelist, limit)
	local names = ''
	if namelist ~= nil then
		if limit == nil then  -- no limit; show all names
			limit = #namelist
		end
		if namelist[1] ~= nil then
			names = namelist[1]
			local maxnames = min(#namelist, limit)
			for i = 2, maxnames do
				if #namelist == 2 then
					names = names .. ' '
				else
					names = names .. ', '
				end
				if i == #namelist then -- last person
					names = names .. 'and '
				end
				names = names .. namelist[i]
			end
			if #namelist > limit then
				names = names .. ', and others'
			end
		end
	end
	return names
end

-- Determines whether item is not an upgrade for the person who looted the item, and is an upgrade for someone else in the group
-- If that's the case, performs the action based on the users' selected Notify Mode
local function PerformNotify(item, characterName)
	local isAnUpgradeForLooter, equippedILVL = IsAnUpgradeForCharacter(item, characterName)
	if equippedILVL > 0 and not isAnUpgradeForLooter then
		-- we now know the item can be traded by the person who received it, so let's check to see if anyone can actually
		--    use the item as an upgrade
		local isAnUpgradeForAnyCharacter, isAnUpgradeForAnyCharacterNames = IsAnUpgradeForAnyCharacter(item)
		if isAnUpgradeForAnyCharacter then
			local names = GetNames(isAnUpgradeForAnyCharacterNames, MAX_NAMES_TO_SHOW)

			if PLH_NOTIFY_MODE == NOTIFY_MODE_GROUP or PLH_NOTIFY_MODE == NOTIFY_MODE_COORDINATE_ROLLS then
				PLH_SendBroadcast(PLH_GetNameWithoutRealm(characterName) .. ' can trade ' .. item .. ', which is an ilvl upgrade for ' .. names)
			end
			if PLH_NOTIFY_MODE == NOTIFY_MODE_COORDINATE_ROLLS and UnitIsGroupLeader('player') then
				if #isAnUpgradeForAnyCharacterNames > 1 then  -- more than 1 person can use the item
					PLH_SendWhisper('You can trade ' .. item .. ', which is an ilvl upgrade for ' .. names .. '. Reply \'' .. TRADE_MESSAGE .. '\' to initiate rolls for this item.', characterName)
					whisperedItems[characterName] = item  -- use full name-realm since that what we'll get when we look it up from the whisper
				end
			end

			if characterName == PLH_GetUnitNameWithRealm('player') then  -- player can trade an item
				PLH_SendAlert('You can trade ' .. item .. ', which is an ilvl upgrade for ' .. names)
				PlaySound('GLUECREATECHARACTERBUTTON')
			elseif IsPlayerInUpgradeList(isAnUpgradeForAnyCharacterNames) then  -- player can receive an item
				PLH_SendAlert(PLH_GetNameWithoutRealm(characterName) .. ' can trade ' .. item .. ', which is an ilvl upgrade for you!')
				PlaySound('LEVELUP')
			elseif PLH_NOTIFY_MODE == NOTIFY_MODE_SELF then
				PLH_SendUserMessage(PLH_GetNameWithoutRealm(characterName) .. ' can trade ' .. item .. ', which is an ilvl upgrade for ' .. names)
			end
		end
	end
end

-- returns true if the item should be evaluated for potential trades based on the following criteria:
--   1. item is equippable
--   2. ilvl is >= min ilvl from preferences
--   3. quality is >= min quality from preferences
--   4. item is BoP, or user specified to include BoE items in preferences
local function ShouldBeEvaluated(item)
	if not IsEquippableItem(item) then
		return false
	else
		local ilvl = PLH_GetRealILVL(item)
		if ilvl < PLH_MIN_ILVL then
			return false
		else
			local quality = select(3, GetItemInfo(item))
			if quality < PLH_MIN_QUALITY then
				return false
			else
				if not PLH_INCLUDE_BOE and not PLH_IsBoundToPlayer(item) then
					return false
				else
					return true
				end
			end
		end
	end
end

local function LootReceivedEvent(self, event, ...)
	local arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 = ...
	local owner
	local lootedItem

	local message = arg1			
	_, _, lootedItem = string.find(message, 'You receive loot: (|.+|r)')
	if lootedItem then
		if ShouldBeEvaluated(lootedItem) then
			PerformNotify(lootedItem, PLH_GetUnitNameWithRealm('player'))
		end
	else
		_, _, owner, lootedItem = string.find(message, '(.+) receives loot: (|.+|r)')

		if lootedItem then
			if ShouldBeEvaluated(lootedItem) then
				owner = PLH_GetUnitNameWithRealm(owner)
				PerformNotify(lootedItem, owner)
			end
		end
	end
end	

local function QueueItem(sender, item)
	queuedRollOwners[numOfQueuedRollItems] = sender
	queuedRollItems[numOfQueuedRollItems] = item
	numOfQueuedRollItems = numOfQueuedRollItems + 1
end

local function AskForRolls()
	if currentRollItem == nil and numOfQueuedRollItems > 0 then
		currentRollOwner = queuedRollOwners[numOfQueuedRollItems - 1]
		currentRollItem = queuedRollItems[numOfQueuedRollItems - 1]
		numOfQueuedRollItems = numOfQueuedRollItems - 1
		PLH_SendBroadcast('Roll for ' .. currentRollItem .. ' from ' .. PLH_GetNameWithoutRealm(currentRollOwner), true)

		local FiveSecondWarningDisplayed = false
		local FifteenSecondWarningDisplayed = false
		
		rollDelayFrame:SetScript('OnUpdate', function(self, elapsed)
			delay = delay + elapsed
--			if currentRollItem == nil then
				-- that means the user forced rolls to end, so we can stop the countdown; relying on PLH_EndRolls() to clean us up
--			elseif delay >= 15 and not FifteenSecondWarningDisplayed then
			if delay >= 15 and not FifteenSecondWarningDisplayed then
				PLH_SendBroadcast('15 seconds remaining to roll for ' .. currentRollItem, false)
				FifteenSecondWarningDisplayed = true
			elseif delay >= 25 and not FiveSecondWarningDisplayed then
				PLH_SendBroadcast('5 seconds remaining to roll for ' .. currentRollItem, false)
				FiveSecondWarningDisplayed = true
			elseif delay >= 30 then
				-- force an end to the rolls
				PLH_EndRolls()
			end
		end)
	end
end

local function ClearRolls()
	currentRollOwner = nil
	currentRollItem = nil
	currentRolls = {}
	queuedRollOwners[numOfQueuedRollItems] = nil
	queuedRollItems[numOfQueuedRollItems] = nil
end

-- not local, because it has to be callable by the global slash command
function PLH_EndRolls()
	-- clear the countdown for the existing roll
	rollDelayFrame:SetScript('OnUpdate', nil)
	delay = 0
	
	-- determine the winner(s)
	local winners = nil
	local highRoll = 0
	for name, roll in pairs(currentRolls) do
		if winners == nil or roll > highRoll then
			winners = { name }
			highRoll = roll
		elseif winners ~= nil and roll == highRoll then
			table.insert(winners, name)
		end
	end
	
	-- notify everyone of the results of the rolls
	if winners ~= nil then
		if #winners > 1 then
			PLH_SendBroadcast(GetNames(winners) .. ' tied for ' .. currentRollItem .. ' from ' .. PLH_GetNameWithoutRealm(currentRollOwner) .. ' with a ' .. highRoll, true)
		else
			PLH_SendBroadcast(winners[1] .. ' won ' .. currentRollItem .. ' from ' .. PLH_GetNameWithoutRealm(currentRollOwner) .. ' with a ' .. highRoll, true)
		end
	else
		PLH_SendBroadcast('Nobody rolled for ' .. currentRollItem .. ' from ' .. PLH_GetNameWithoutRealm(currentRollOwner), true)
	end

	-- if there's another roll to do, let's delay a few seconds before starting the next roll
	if (numOfQueuedRollItems > 0) then
		rollDelayFrame:SetScript('OnUpdate', function(self, elapsed)
			nextRollDelay = nextRollDelay + elapsed
			if nextRollDelay >= DELAY_BETWEEN_ROLLS then
				rollDelayFrame:SetScript('OnUpdate', nil)
				nextRollDelay = 0
				ClearRolls()
				AskForRolls()
			end
		end)
	else			
		ClearRolls()
	end	
end

local function GetItemFromQueueByPlayer(player)
	for key, value in pairs(queuedRollOwners) do
		if value == player then
			return queuedRollItems[key]
		end
	end
	return nil
end

local function WhisperReceivedEvent(self, event, ...)
	local arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12 = ...
	local message = arg1
	local sender = arg2

	if string.upper(message) == TRADE_MESSAGE or string.upper(message) == '\'' .. TRADE_MESSAGE .. '\'' then
		if whisperedItems[sender] ~= nil then
			local item = whisperedItems[sender]
			whisperedItems[sender] = nil
			QueueItem(sender, item)
			-- if we're still rolling for another item, let the person know their item is queued
			if currentRollItem ~= nil then
				PLH_SendWhisper('Thank you! ' .. item .. ' will be rolled for after current rolls are done.', sender)
			else 
				AskForRolls()
			end
		elseif currentRollOwner == sender then
			PLH_SendWhisper('Your ' .. currentRollItem .. ' is currently being rolled for', sender)
		else
			local item = GetItemFromQueueByPlayer(sender)
			if item then
				PLH_SendWhisper('Your ' .. item .. ' is already in queue to be rolled for', sender)
			else
				PLH_SendWhisper('No record of which item you looted! Whisper the item link to ' .. UnitName('player') .. ' if you would still like to trade the item.', sender)
			end
		end
	end
end

local function RollReceivedEvent(self, event, ...)
	if currentRollItem ~= nil then
		local message = select(1, ...)
		if message then
			local name, roll, minRoll, maxRoll = message:match('^(.+) rolls (%d+) %((%d+)%-(%d+)%)$')
			if name then
				if minRoll ~= '1' or maxRoll ~= '100' then
					PLH_SendBroadcast(name .. ' rolled ' .. minRoll .. ' - ' .. maxRoll .. '; roll ignored', false)
				elseif currentRolls[name] ~= nil then
					PLH_SendBroadcast(name .. ' rolled multiple times; only the first roll of ' .. currentRolls[name] .. ' counts', false)
				elseif not IsEquippableItemForCharacter(currentRollItem, name) then
					PLH_SendBroadcast(name .. ' is not eligible for ' .. currentRollItem .. '; roll ignored', false)
				else 
					currentRolls[name] = roll
				end
			end		
		end
	end
end

-- note that GetLootMethod() only works if you're in a party or in a raid.  If you're in an instance (i.e. queued via LFR),
--   then loot method is personal loot
local function IsPersonalLoot()
	local method, _, _ = GetLootMethod()
	return IsInInstance() or method == 'personalloot'
end

local function ShouldBeEnabled()
	return (IsPersonalLoot() and IsInGroup())
end

local function ResetVariables()
	delay = 0
	nextRollDelay = 0

	whisperedItems = {}

	numOfQueuedRollItems = 0;
	queuedRollOwners = {}
	queuedRollItems = {}

	currentRollOwner = nil
	currentRollItem = nil
	currentRolls = {}

	groupInfoCache = {}
end

-- The following uses GetInventoryItemLink() to look up unit's equipped items.  That method can only be called for
--   the player, or within the scope of an INSPECT_READY event for other group members.
local function UpdateGroupInfoCache(unit)
	local name = PLH_GetUnitNameWithRealm(unit)

	if name ~= nil then
--		PLH_SendDebugMessage('   Updating GroupInfoCache for ' .. name .. ', inspectIndex = ' .. inspectIndex)
		local characterDetails
		if groupInfoCache[name] == nil then
			characterDetails = {}
			local class = UnitClass(unit)
			characterDetails['ClassName'] = class
			local spec = GetInspectSpecialization(unit)
			characterDetails['Spec'] = spec
			local level = UnitLevel(unit)
			characterDetails['Level'] = level
		else
			characterDetails = groupInfoCache[name]
			spec = GetInspectSpecialization(unit)
		end
		
		local updatedItemCount = 0
		local item
		for i = _G.INVSLOT_FIRST_EQUIPPED, _G.INVSLOT_LAST_EQUIPPED do
			if i ~= _G.INVSLOT_BODY and i ~= INVSLOT_TABARD then -- ignore shirt and tabard slots
				item = GetInventoryItemLink(UnitName(unit), i)
				if item ~= nil then
					if characterDetails[i] == nil or characterDetails[i] ~= item then
--					PLH_SendDebugMessage('      adding/updating item ' .. item)			
						updatedItemCount = updatedItemCount + 1
						characterDetails[i] = nil  -- remove existing first
						characterDetails[i] = item
					end
					
					-- if we're adding the weapon, also add relics
					if i == _G.INVSLOT_MAINHAND or i == _G.INVSLOT_OFFHAND then
						local _, relic1 = GetItemGem(item, 1)
						if relic1 ~= nil then
							characterDetails[PLH_RELICSLOT + 1] = relic1
						end
						local _, relic2 = GetItemGem(item, 2)
						if relic2 ~= nil then
							characterDetails[PLH_RELICSLOT + 2] = relic2
						end
						local _, relic3 = GetItemGem(item, 3)
						if relic3 ~= nil then
							characterDetails[PLH_RELICSLOT + 3] = relic3
						end
					end
				end
			end
		end

		-- If we didn't find any items (ex: when player first starts the game), don't add this person to the cache
		if updatedItemCount > 0 then
			groupInfoCache[name] = nil  -- remove any existing entry first
			groupInfoCache[name] = characterDetails
		end
		PLH_SendDebugMessage('   Updated ' .. updatedItemCount .. ' items for ' .. name .. '; loop ' .. inspectLoop .. '; char ' .. (inspectIndex - 1))
	end
end

-- returns true if the characterName is in the raid/party
local function IsCharacterInGroup(characterFullname)
	local index = 1
	local name = select(1, GetRaidRosterInfo(index))
	while name ~= nil do
		if name == characterFullname then
			return true
		end
		-- RaidRosterInfo will only give us name by itself if the character is on the same realm as the player, so check that scenario too
		if PLH_GetFullName(name, GetRealmName()) == characterFullname then
			return true
		end
		index = index + 1
		name = select(1, GetRaidRosterInfo(index))
	end
	return false
end

-- returns true if group member was able to be inspected; false otherwise
local function InspectGroupMember(characterName)
	if characterName ~= nil and characterName ~= UnitName('player') then   -- no need to inspect ourselves
		if CanInspect(characterName) then -- and UnitIsConnected(characterName) - not necessary to include this check since we're including distance check
--			if CheckInteractDistance(characterName, 1) then  -- this API call has not been changed by Blizz to match the new larger inspect radius
			if UnitIsVisible(characterName) then
				NotifyInspect(characterName)
				notifyInspectName = characterName
				PLH_wait(DELAY_BETWEEN_INSPECTIONS, PLH_InspectNextGroupMember)
				return true
			else
				PLH_SendDebugMessage('   ' .. characterName .. ' out of range for inspect')
			end
		else
			PLH_SendDebugMessage('   Unable to inspect ' .. characterName)
		end
	end
	return false
end

-- An Inspect Loop (managed by inspectLoop) is a complete iteration of inspection for every member in the group.  
-- 		Will only attempt to inspect characters whose count of cached items is lower than expected.
--      The goal of this loop is to work around the limitation whereby the inspect API doesn't necessarily provide us
--      all items equipped by the character.
-- not local, because it is called by (and calls) InspectGroupMember(characterName), which is defined above
function PLH_InspectNextGroupMember()
	-- attempt the inspect the next person; if they're not inspectable, move onto the next

	-- Removed this inside-loop retry logic.  It works, but in testing it very rarely accomplished its goal; usually the retry attempts
	--    to inspect failed just like the original attempt.  We now have inspectLoop logic build in, which will still
	--    give the cache more opportunities to be updated - so no need for the retries here
	-- Retry logic
--	if notifyInspectName ~= nil then   -- InspectReady didn't get called for the person we queued
--		if inspectRetries < MAX_INSPECT_RETRIES then
--			PLH_SendDebugMessage('Retrying inspection for ' .. notifyInspectName)
--			inspectRetries = inspectRetries + 1
--			if InspectGroupMember(notifyInspectName) then  -- we triggered a notify, so don't do the while loop
--				return true  -- exit this function so we don't attempt to inspect the next character
--			end
--		end
		-- if we get to here, then it's time to give up on inspecting this character - either we couldn't even trigger a
		--    NotifyInspect, or we already triggered the max # of NotifyInspects and didn't get an InspectReady response
		notifyInspectName = nil  -- even though we commented out the retry logic, we still want to nullify notifyInspectName
			-- since it's unlikely we'll get an INSPECT_READY at this point (3 seconds after calling NotifyInspect)
--		inspectRetries = 0
--	end

	-- Call InspectGroupMember for the next person in the group
	local characterName
	local queuedAnInspection = false
	local expectedItemCount
	local expectedRelicCount
	local class
	local spec
	local level
	while inspectIndex <= maxInspectIndex  and not queuedAnInspection do
		characterName = select(1, GetRaidRosterInfo(inspectIndex))
		if characterName ~= nil then  -- safeguard; character may have left the roster between the time we started the call and now
			local numCachedItems = 0
			local numCachedRelics = 0
			local fullname = characterName 		--characterName may or may not have realm.  we want to preserve it the way it is for the call to InspectGroupMember,
												--   but need the name-realm version of the name to look up the element in the cache
			if not string.find(fullname, '-') then
				fullname = PLH_GetFullName(characterName, GetRealmName())
			end
			
			expectedItemCount = NUM_EXPECTED_ITEMS
			expectedRelicCount = 0
			if groupInfoCache[fullname] ~= nil then
				numCachedItems = PLH_GetItemCountFromCache(fullname)
				numCachedRelics = PLH_GetRelicCountFromCache(fullname)
				class = groupInfoCache[fullname]['ClassName']
				spec = groupInfoCache[fullname]['Spec']
				level = groupInfoCache[fullname]['Level']
				expectedItemCount = GetExpectedItemCount(class, spec, level)
				expectedRelicCount = GetExpectedRelicCount(level)
			end

			if numCachedItems < expectedItemCount or numCachedRelics < expectedRelicCount then  -- if we've already cached 15 or more items, don't bother refreshing
				queuedAnInspection = InspectGroupMember(characterName)
			end
		end
		inspectIndex = inspectIndex + 1
	end
	
	-- The following logic is meant to work around a limitation of the inspect API.  When you inspect a character, you're
	-- not guaranteed to actually receive all of their equipped items back!  To work around this limitation, we will
	-- perform additional loops of inspecting each character if the number of items we've cached for them is fewer than
	-- the expected number of items that someone would equip (15 items for a person using a 1-hander, +1 element for the
	-- ClassName that we're storing in the cache)
	if inspectIndex > maxInspectIndex then -- that means we just completed our current loop
		if inspectLoop < MAX_INSPECT_LOOPS then
			-- let's start the next loop
			inspectIndex = 1  -- we're triggering the call to the 1st member via PLH_InspectNextGroupMember below; increment
				-- the inspectIndex counter so the next element that gets picked up when we come back into PLH_InspectNextGroupMember
				-- is the 2nd
			inspectLoop = inspectLoop + 1
			if queuedAnInspection then  -- if we just queued someone for inspection, wait before we start the new loop
				PLH_wait(DELAY_BETWEEN_INSPECTIONS, PLH_InspectNextGroupMember)
			else  -- otherwise start the new loop immediately
				PLH_InspectNextGroupMember()
			end
		elseif not queuedAnInspection then
			-- we've finished all loops
--			PLH_PrintCache()
		end
	end
	
end

-- note that the INSPECT_READY event may have been triggered by WoW or by another addon
--   for example, WoW triggers it when a person in inspect range swaps their gear
local function InspectReadyEvent(self, event, ...)
	local guid = select(1, ...)
	local _, _, _, _, _, name, realm = GetPlayerInfoByGUID(guid)

	if notifyInspectName ~= nil and (notifyInspectName == name or notifyInspectName == PLH_GetFullName(name, realm)) then  -- confirm this InspectReadyEvent is for the person we triggered for inspection.
		UpdateGroupInfoCache(name)  -- only update the cache if we requested the inspect, otherwise the wrong cache entry could be updated!
		-- if we triggered the inspection, clear the flags that we had set
		ClearInspectPlayer()  -- note that we only want to call ClearInspectPlayer() if we're the ones who queued up
			-- the inspection; otherwise, we'll prevent whoever wanted to inspect the player from seeing data...for
			-- example, a player using the UI to inpect a player would see all empty slots if we called ClearInspectPlayer()!
		notifyInspectName = nil  -- clear this flag since we're done inspecting the person
	end
end

local function PopulateGroupInfoCache()
	local now = time()
	
	if now - priorCacheRefreshTime > 3 then  -- only refresh if prior refresh was more than 3 seconds ago
		priorCacheRefreshTime = now
	
		-- remove characters from the cache if they're no long in the raid/party
		for name, details in pairs (groupInfoCache) do
			if not IsCharacterInGroup(name) then
				PLH_SendDebugMessage('Removing entry for ' .. name .. ' from cache')
				groupInfoCache[name] = nil
			end
		end

		if IsInGroup() then
			-- If we're already doing an inspect loop, don't interupt it; just do nothing with the request to
			-- PopulateGroupInfoCache() and let the inspect loop continuue on its way!  If the inspectIndex > maxInspectIndex
			-- and inspectLoop > MAX_INSPECT_LOOPS, then we know we've finished all inspections for all loops, so
			-- we can start a brand new loop!
			if inspectIndex > maxInspectIndex and inspectLoop >= MAX_INSPECT_LOOPS then
--				PLH_SendDebugMessage('Refreshing cache')
				inspectIndex = 1
				inspectLoop = 1
				maxInspectIndex = GetNumGroupMembers()
				PLH_InspectNextGroupMember()
			end
		end
	end
end

local function Enable()
	ResetVariables()
	isEnabled = true
	lootReceivedEventFrame:RegisterEvent('CHAT_MSG_LOOT')
	whisperReceivedEventFrame:RegisterEvent('CHAT_MSG_WHISPER')
	rollReceivedEventFrame:RegisterEvent('CHAT_MSG_SYSTEM')
	inspectReadyEventFrame:RegisterEvent('INSPECT_READY')
	combatStatusChangedEventFrame:RegisterEvent('PLAYER_REGEN_DISABLED')   -- player entered combat
	groupMemberInfoChangedEventFrame:RegisterEvent('PLAYER_SPECIALIZATION_CHANGED')
	groupMemberInfoChangedEventFrame:RegisterEvent('UNIT_INVENTORY_CHANGED')
	PLH_SendStatusMessage(false)
end

local function Disable()
	isEnabled = false
	lootReceivedEventFrame:UnregisterAllEvents()
	whisperReceivedEventFrame:UnregisterAllEvents()
	rollReceivedEventFrame:UnregisterAllEvents()
	inspectReadyEventFrame:UnregisterAllEvents()
	combatStatusChangedEventFrame:UnregisterAllEvents()
	groupMemberInfoChangedEventFrame:UnregisterAllEvents()
	groupInfoCache = {}
	PLH_SendStatusMessage(false)
end

local function EnableOrDisable()
	local shouldBeEnabled = ShouldBeEnabled()
	if not isEnabled and shouldBeEnabled then	
		Enable()
	elseif isEnabled and not shouldBeEnabled then
		Disable()
	end
	if isEnabled then 
		PopulateGroupInfoCache()
	end
end

local function GroupMemberInfoChangedEvent(self, event, ...)
	local unit = ...
	if unit == 'player' then
		-- do nothing
	else
		-- update cache
		local name = PLH_GetUnitNameWithRealm(unit)
		groupInfoCache[name] = nil
		if event == 'UNIT_INVENTORY_CHANGED' then
			-- we need to do another inspect to refresh the cache
			PopulateGroupInfoCache()  -- don't want to mess with any existing cache refresh flow by inspecting this specific character,
				-- so trigger general refresh, which will refresh this character since we've nullified their entry
		else  -- PLAYER_SPECIALIZATION_CHANGED
			-- spec change already gives us the inspect info
			UpdateGroupInfoCache(unit)
		end
	end
end

local function RosterUpdatedEvent(self, event, ...)
	-- the following is a bit of a hack to work around a Blizzard issue.  While the player is logging in, IsInGroup()
	-- is false.  If the user is already in a group (for example, logging back in after a disconnect or doing a /reload),
	-- A ROSTER_UPDATE event triggers.  However, IsInGroup() is not immediately set to true when the event fires!
	-- So if we get a ROSTER_UPDATE event and we're currently disabled, lets wait 1 second to make sure IsInGroup()
	-- gives us the correct value.  Similar behavoir occurred in LFR testing where people joining/leaving the group
	-- may not have been automatically available.  Hence the delay.
	PLH_wait(2, EnableOrDisable)
end

-- triggered when the player enters or leaves combat status, which are the perfect times to refresh the cache since
--    the people who will be eligible for loot should be close enough to be inspected
local function CombatStatusChangedEvent(self, event, ...)
	if isEnabled then
		PopulateGroupInfoCache()
	end
end

function SlashCmdList.PLHelperCommand(msg, editbox)
	if msg == nil or msg == '' then
		InterfaceOptionsFrame_OpenToCategory(PLH_LONG_ADDON_NAME)
	elseif msg == 'endroll' then
		if currentRollItem ~= nil then
			if nextRollDelay > 0 then
				PLH_SendUserMessage('Please wait until the next roll begins')
			else
--				PLH_SendBroadcast(UnitName('player') .. ' ended rolls', false)
				PLH_EndRolls()
			end
		else
			PLH_SendUserMessage('There are currently no items being rolled for')
		end
--	elseif msg == 'test' then
--		PLH_UnitTest()
	else
		PLH_SendUserMessage('Unknown parameter. Options are:\n  [/plh]  :  open interface options\n  [/plh endroll]  :  force current roll to end')
	end
end

function PLH_SendStatusMessage(onlyShowIfEnabled)
	if isEnabled then
		if PLH_NOTIFY_MODE == NOTIFY_MODE_COORDINATE_ROLLS then
			PLH_SendUserMessage('Enabled in \'Notify Group + Coordinate Rolls\' mode. Please turn off \'Notify Group\' if someone else in the group is already running PLH in \'Notify Group\' mode. Mode can be changed through interface options.')
		elseif PLH_NOTIFY_MODE == NOTIFY_MODE_GROUP then
			PLH_SendUserMessage('Enabled in \'Notify Group\' mode. Please turn off \'Notify Group\' if someone else in the group is already running PLH in \'Notify Group\' mode. Mode can be changed through interface options.')
		else
			PLH_SendUserMessage('Enabled in \'Notify Self\' mode. Mode can be changed through interface options.')
		end
		if PLH_DEBUG then
			PLH_SendUserMessage('Debug mode is enabled. \'/run PLH_DisableDebug()\' to turn debug off.')
		end
	elseif not onlyShowIfEnabled then
		PLH_SendUserMessage('Disabled')
	end
end

local function InitializeOptions()
	if PLH_NOTIFY_MODE == nil then
		PLH_NOTIFY_MODE = DEFAULT_NOTIFY_MODE
	end
	if PLH_INCLUDE_BOE == nil then
		PLH_INCLUDE_BOE = DEFAULT_INCLUDE_BOE
	end
	if PLH_MIN_ILVL == nil then
		PLH_MIN_ILVL = DEFAULT_MIN_ILVL
	end
	if PLH_MIN_QUALITY == nil then
		PLH_MIN_QUALITY = DEFAULT_MIN_QUALITY
	end
	if PLH_DEBUG == nil then
		PLH_DEBUG = DEFAULT_DEBUG
	end
	if PLH_CURRENT_SPEC_ONLY == nil then
		PLH_CURRENT_SPEC_ONLY = DEFAULT_CURRENT_SPEC_ONLY
	end
end

local function Initialize(self, event, addonName, ...)
	if addonName == 'PersonalLootHelper' then
		InitializeOptions()

		if lootReceivedEventFrame == nil then
			lootReceivedEventFrame = CreateFrame('Frame')
			lootReceivedEventFrame:SetScript('OnEvent', LootReceivedEvent)

			whisperReceivedEventFrame = CreateFrame('Frame')
			whisperReceivedEventFrame:SetScript('OnEvent', WhisperReceivedEvent)

			rollReceivedEventFrame = CreateFrame('Frame')
			rollReceivedEventFrame:SetScript('OnEvent', RollReceivedEvent)

			inspectReadyEventFrame = CreateFrame('Frame')
			inspectReadyEventFrame:SetScript('OnEvent', InspectReadyEvent)

			rosterUpdatedEventFrame = CreateFrame('Frame')
			rosterUpdatedEventFrame:SetScript('OnEvent', RosterUpdatedEvent)
			rosterUpdatedEventFrame:RegisterEvent('GROUP_ROSTER_UPDATE')
			
			groupMemberInfoChangedEventFrame = CreateFrame('Frame')
			groupMemberInfoChangedEventFrame:SetScript('OnEvent', GroupMemberInfoChangedEvent)
			
			combatStatusChangedEventFrame = CreateFrame('Frame')
			combatStatusChangedEventFrame:SetScript('OnEvent', CombatStatusChangedEvent)

			rollDelayFrame = CreateFrame('Frame')
		end
	end
end

-- Initialize()

local initializeFrame = CreateFrame('Frame')
initializeFrame:SetScript('OnEvent', Initialize)
initializeFrame:RegisterEvent('ADDON_LOADED')

--[[
*********************************************************
Debug/Testing functions
*********************************************************
]]--

function PLH_PrintCache(showDetails, characterName)
	if PLH_DEBUG then
		local num_characters = 0
		local item_msg = ''
		for name, characterDetails in pairs(groupInfoCache) do
			num_characters = num_characters + 1
			item_msg = item_msg .. PLH_GetItemCountFromCache(name) .. '/' .. PLH_GetRelicCountFromCache(name) .. ' ' 
--			item_msg = item_msg .. (#groupInfoCache[name] -1) .. ' '  -- subtracting 1 since everyone will have a ClassName element
		end
		PLH_SendDebugMessage('Cache contains ' .. num_characters .. ' member(s). Item count per member: ' .. item_msg)

		if (showDetails) then
			for name, details in pairs (groupInfoCache) do
				if name == nil or name == characterName then
					PLH_SendDebugMessage('Cache information for ' .. name)
					if details == nil then	
						PLH_SendDebugMessage('   details is nil')
					else
						for slotID, item in pairs(details) do
							PLH_SendDebugMessage('   ' .. slotID .. ' = ' .. item)
						end
					end
				end
			end
		end
	end
end

function PLH_EnableDebug()
	PLH_DEBUG = true
end

function PLH_DisableDebug()
	PLH_DEBUG = false
end

function PLH_RefreshCache()
	groupInfoCache = {}
	PopulateGroupInfoCache()
end

function PLH_TestItems(characterIndex)
	if characterIndex == nil then
		PLH_SendDebugMessage('Usage: PLH_TestItems(characterIndex)')
	else
		characterName = select(1, GetRaidRosterInfo(characterIndex))
		if not string.find(characterName, '-') then
			characterName = PLH_GetFullName(characterName, GetRealmName())
		end
		
		PLH_SendDebugMessage('Evaluating items equipped by ' .. characterName)

		local item
		for itemIndex = 1, 19 do
			if characterName == PLH_GetUnitNameWithRealm('player') then
				item = GetInventoryItemLink('player', itemIndex)	
			else
				item = groupInfoCache[characterName][itemIndex]
			end

			if item ~= nil then
				PLH_SendDebugMessage('   evaluating ' .. item)
				
				local isEquippable
				for evalIndex = 1, GetNumGroupMembers() do
					evalName = select(1, GetRaidRosterInfo(evalIndex))
					if not string.find(evalName, '-') then
						evalName = PLH_GetFullName(evalName, GetRealmName())
					end
				
					isEquippable = IsEquippableItemForCharacter(item, evalName)
					PLH_SendDebugMessage('      For ' .. evalName ..
						' equippable = ' .. tostring(isEquippable) ..
						'; upgrade = ' .. tostring(isEquippable and IsAnUpgradeForCharacter(item, evalName))
						)
				end
			end
		end
		
		local relic
		for relicIndex = 1, 3 do
			if characterName == PLH_GetUnitNameWithRealm('player') then
				local weapon = GetInventoryItemLink('player', INVSLOT_MAINHAND)
				_, relic = GetItemGem(weapon, relicIndex)
			else
				relic = groupInfoCache[characterName][PLH_RELICSLOT + relicIndex]
			end

			if relic ~= nil then
				PLH_SendDebugMessage('   evaluating ' .. relic)
				
				local isEquippable
				for evalIndex = 1, GetNumGroupMembers() do
					evalName = select(1, GetRaidRosterInfo(evalIndex))
					if not string.find(evalName, '-') then
						evalName = PLH_GetFullName(evalName, GetRealmName())
					end
				
					isEquippable = IsEquippableItemForCharacter(relic, evalName)
					PLH_SendDebugMessage('      For ' .. evalName ..
						' equippable = ' .. tostring(isEquippable) ..
						'; upgrade = ' .. tostring(isEquippable and IsAnUpgradeForCharacter(relic, evalName))
						)
				end
			end
		end
	end
end

-- note that some classes can have 2 of the same relic; take that into account
--[[
when caching, cache RelicSlot1, RelicSlot2, RelicSlot3
add IsArtifactRelicItem to IsEquippableItem() check(s)
if class = Gem and subclass = Artifact Relic, branch off into specific logic
for 1 to 3
	if socket type[i] is same as type of looted relic then
		compare ilvls
need function to get socket type given a relic (parse tooltip looking for " Artifact Relic")
need constant that indicates socket type by slot for each class/spec
]]--
function PLH_TestRelic(relic, item)
	local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(relic)
	print('name ', name)
	print('class ', class)  -- 'Gem'
--	print('iLevel ', iLevel)  -- gives wrong value
	print('subclass ', subclass)  -- 'Artifact Relic'
--	print('equipSlot ', equipSlot)  -- nil
	print('real ilvl is ', PLH_GetRealILVL(relic))  -- correct
	print('IsEquippableItem	is ', IsEquippableItem(relic))  -- false, so we'll need to change where we do this check!
	print('IsArtifactRelicItem is ', IsArtifactRelicItem(relic))  -- true; can add this where we check isEquippable
	
	local gemname, gemlink = GetItemGem(item, 1)  -- 2nd arg is index; does return gem link from artifact
	print('gemname ', gemname)
	print('gemlink ', gemlink)
	print('real ilvl is ', PLH_GetRealILVL(gemlink))  -- correct

	local gemname, gemlink = GetItemGem(item, 2)  -- 2nd arg is index; does return gem link from artifact
	print('gemname ', gemname)
	print('gemlink ', gemlink)
	print('real ilvl is ', PLH_GetRealILVL(gemlink))  -- correct

	local gemname, gemlink = GetItemGem(item, 3)  -- 2nd arg is index; does return gem link from artifact
	print('gemname ', gemname)
	print('gemlink ', gemlink)
	print('real ilvl is ', PLH_GetRealILVL(gemlink))  -- correct

--	local slotId, texture, checkRelic = GetInventorySlotInfo("MainHandSlot")
--	print('checkRelic1 ', checkRelic)  -- false
	
--	local slotId, texture, checkRelic = GetInventorySlotInfo("RangedSlot")
--	print('checkRelic2 ', checkRelic)  -- invalid slot error

	-- following method not found even though it's in API docs
--	local gem1, gem2, gem3 = GetInventoryItemGems(INVSLOT_RANGED)	--18
--	local gem1, gem2, gem3 = GetInventoryItemGems(18)	--18
--	print('gem1 ', gem1)
--	print('gem2 ', gem2)
--	print('gem3 ', gem3)
end

--[[
*********************************************************
Code from BlizzBugsSuck.lua to resolve known blizzard bug when opening interface options
*********************************************************
]]--
-- Fix InterfaceOptionsFrame_OpenToCategory not actually opening the category (and not even scrolling to it)
-- Confirmed still broken in 6.2.2.20490 (6.2.2a)
do
	local function get_panel_name(panel)
		local tp = type(panel)
		local cat = INTERFACEOPTIONS_ADDONCATEGORIES
		if tp == "string" then
			for i = 1, #cat do
				local p = cat[i]
				if p.name == panel then
					if p.parent then
						return get_panel_name(p.parent)
					else
						return panel
					end
				end
			end
		elseif tp == "table" then
			for i = 1, #cat do
				local p = cat[i]
				if p == panel then
					if p.parent then
						return get_panel_name(p.parent)
					else
						return panel.name
					end
				end
			end
		end
	end

	local function InterfaceOptionsFrame_OpenToCategory_Fix(panel)
		if doNotRun or InCombatLockdown() then return end
		local panelName = get_panel_name(panel)
		if not panelName then return end -- if its not part of our list return early
		local noncollapsedHeaders = {}
		local shownpanels = 0
		local mypanel
		local t = {}
		local cat = INTERFACEOPTIONS_ADDONCATEGORIES
		for i = 1, #cat do
			local panel = cat[i]
			if not panel.parent or noncollapsedHeaders[panel.parent] then
				if panel.name == panelName then
					panel.collapsed = true
					t.element = panel
					InterfaceOptionsListButton_ToggleSubCategories(t)
					noncollapsedHeaders[panel.name] = true
					mypanel = shownpanels + 1
				end
				if not panel.collapsed then
					noncollapsedHeaders[panel.name] = true
				end
				shownpanels = shownpanels + 1
			end
		end
		local Smin, Smax = InterfaceOptionsFrameAddOnsListScrollBar:GetMinMaxValues()
		if shownpanels > 15 and Smin < Smax then
			local val = (Smax/(shownpanels-15))*(mypanel-2)
			InterfaceOptionsFrameAddOnsListScrollBar:SetValue(val)
		end
		doNotRun = true
		InterfaceOptionsFrame_OpenToCategory(panel)
		doNotRun = false
	end

	hooksecurefunc("InterfaceOptionsFrame_OpenToCategory", InterfaceOptionsFrame_OpenToCategory_Fix)
end
--[[
-- following only returns for some items (notably trinkets), and only returns spec IDs relevant to the player
function PrintItemSpecInfo(item)
	specs = {}
	results = GetItemSpecInfo(item, specs)
	print('results is ', results)
	print('#results is ', #results)
	print('#specs is ', #specs)
	for i = 1, #results do
		print(results[i])
	end
	for i = 1, #specs do
		print(results[i])
	end
	for key, value in pairs(results) do
		print('key = ', key)
		print('value = ', value)
	end
	for key, value in pairs(specs) do
		print('key2 = ', key)
		print('value2 = ', value)
	end
end
]]--